==> bordeaux_0.4.6.orig.tar.gz: bordeaux-0.4.6/server/main.cpp <==
/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.hpp"
#include "utils.hpp"
#include "ipc.hpp"
#include "logger.hpp"
#include "network.hpp"
#include "drivers.hpp"
#include "sessions.hpp"
#include "proc.hpp"

#include <args.hpp>
#include <keyfile.hpp>
#include <strings.hpp>

#include <atomic>
#include <thread>
#include <chrono>
#include <fstream>
#include <iostream>
#include <csignal>
#include <cstdio>
#include <new>

#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pwd.h>
#include <fcntl.h>
#include <dlfcn.h>

#ifndef BUILD_DOCKER
#include <mqueue.h>
#endif

using namespace server;

std::atomic<bool> server::running = false;  // need not be atomic...but...
unsigned server::logging = 0;
bool server::ipv6 = false;
uid_t server::nobody = 65534;

namespace {
auto driver_name = BORDEAUX_DEFAULT_DRIVER;
const char *config_file = SERVER_CONFIG_FILE;
const char *prefix_path = SERVER_PREFIX_PATH;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config(0,  "--config", "configuration file", "file");
args::option opt_driver('d',"--driver", "driver name to load", "path");
args::option helpflag('h',  "--help", "display this list");
args::option paths(0,       "--paths", "display paths used");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);

auto reload()
{
    keyfile config({config_file, "custom.conf"});
    auto& hostname = config["server"]["hostname"];
    auto& username = config["server"]["username"];
    auto& media = config["server"]["media"];
    auto& language = config["server"]["language"];
    auto& voice = config["server"]["voice"];
    auto& ring_timer = config["timers"]["ring"];
    auto& answer_timer = config["timers"]["answer"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    if(hostname.empty()) {
        auto env = util::get_env("BORDEAUX_HOSTNAME", 255);
        if(!env || !*env)
            env = util::get_env("PUBLIC_HOSTNAME", 255);

        if(env && *env)
            hostname = env;
        else
            hostname = net::hostname();
    }

    if(language.empty()) {
        auto env = util::get_env("BORDEAUX_LANGUAGE", 16);
        if(!env || !*env)
            env = util::get_env("LANGUAGE", 16);

        if(!env || !*env)
            env = "en_US";

        language = env;
    }

    if(voice.empty()) {
        auto env = util::get_env("BORDEAUX_VOICE", 64);
        if(!env || !*env)
            env = util::get_env("VOICE", 64);

        if(!env || !*env)
            env = "default";

        voice = env;
    }

    setenv("LANGUAGE", language.c_str(), 1);
    setenv("VOICE", voice.c_str(), 1);

    auto env = util::get_env("BORDEAUX_SESSIONS", 16);
    if(env && *env)
        config["server"]["limit"] = env;

    if(hostname.empty())
        hostname = net::hostname();

    if(ring_timer.empty())
        ring_timer = "4";

    if(answer_timer.empty())
        answer_timer = "30";

    if(!media.empty())
        media = net::uri_address(media);
    else
        media = net::uri_address(hostname);

    net::fqdn(hostname);

    if(username.empty()) {
        auto pwd = getpwuid(geteuid());
        if(pwd)
            username = pwd->pw_name;
        else
            username = "bordeaux";
        endpwent();
    }

    running = true;
    Plugin::config(config);
    return Driver::config_instance(config);
}

auto stop()
{
    auto ctrl = ::open(SERVER_CONTROL_IPC, O_WRONLY);   // FlawFinder: ignored
    if(ctrl != -1) {
        // TODO: Will use special stop token in the future
        ::write(ctrl, "stop\n", 5);
        ::close(ctrl);
    }

    running = false;
    std::this_thread::sleep_for(std::chrono::milliseconds(50));

    Driver::drop_instance();
    Plugin::stop();
    Session::stop();
    Driver::stop_instance();
    ::remove(SERVER_CONTROL_IPC);
}

#ifndef BUILD_DOCKER
void process_events(const ipc_event& event)
{
    if(event.pid == getpid() || event.pid == 0) {
        log_error("ipc: invalid pid");
        return;
    }

    switch(event.type) {
    case IPC_STATUS:
        if(running)
            proc::ipc_notify(event.pid, true);
        else
            proc::ipc_notify(event.pid, false);
        break;
    case IPC_LEVEL:
        logging = event.body.level;
        kill(event.pid, SIGUSR1);
        log_notice("logging level changed to {}", logging);
        break;
    case IPC_RELOAD:
        raise(SIGHUP);
        kill(event.pid, SIGUSR1);
        break;
    default:
        debug(4, "received unknown message {} from pid {}", event.type, event.pid);
    }
}
#endif

void process_control(int priority)
{
    using namespace std;

    // Only meant for server libexec, not group privileged users
    remove(SERVER_CONTROL_IPC);
    if(mkfifo(SERVER_CONTROL_IPC, 0600)) {
        remove(SERVER_CONTROL_IPC);
        mkfifo(SERVER_CONTROL_IPC, 0600);
    }

    fstream in(SERVER_CONTROL_IPC, ios::in|ios::out|ios::app);

    if(!in.is_open()) {
        log_error("control: {}: cannot access", SERVER_CONTROL_IPC);
        return;
    }

    if(!proc::set_priority(priority))
        priority = 0;

    log_info("starting control thread; pri={}", priority);
    std::string line;
    while(running && getline(in, line)) {
        debug(2, "control: {}", line);

        auto list = tokenize<std::string_view>(line);
        if(list.empty()) {
            log_warn("control: empty string");
            continue;
        }

        // TODO: will use random generated local token in future
        if(list[0] == "stop")
            break;

        Driver::control(list);
    }
    log_info("stopping control thread");
}

auto loader(const std::string& prefix, const std::string& name, const std::string& type)
{
    auto module = type;
    module += name;
    module += ".so";
    auto dsopath1 = prefix + "/" + module;
    auto dsopath2 = prefix + "/" + name + "/" + module;
    auto ptr = dlopen(dsopath1.c_str(), RTLD_NOW | RTLD_GLOBAL);
    if(ptr)
        log_info("loading {}", dsopath1);
    else {
        ptr = dlopen(dsopath2.c_str(), RTLD_NOW | RTLD_GLOBAL);
        if(ptr)
            log_info("loading {}", dsopath2);
    }
    if(!ptr)
        log_error("{}: failed to load", name);
    return ptr != nullptr;
}
} // anon namespace

auto main(int argc, const char **argv) -> int   // NOLINT
{
#ifdef  DEBUG
    remove(SERVER_LOGGER_PATH "/bordeaux.log");
#endif

    auto exit_code = 0;
    log_reset();

    atexit([]{
        if(running) {
            stop();
            log_warn("exiting server");
        }
#ifndef BUILD_DOCKER
        mq_unlink(BORDEAUX_IPCPATH);
#endif
        shm_unlink(BORDEAUX_IPCPATH);
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                SERVER_ID " " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Bordeaux Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = *opt_prefix;

        if(is(version)) {
            print("Version: {}\n", PROJECT_VERSION);
            return 0;
        }

        if(is(paths)) {
            print("Config:  {}\nPrefix:  {}\nDrivers: {}\nPlugins: {}\nLibexec: {}\nModules: {}\nScripts: {}\nPrompts: {}\nVoices:  {}\n",
                config_file, prefix_path,
                BORDEAUX_DRIVERS_PATH, BORDEAUX_PLUGINS_PATH,
                BORDEAUX_LIBEXEC_PATH,
                BORDEAUX_MODULES_PATH, BORDEAUX_SCRIPTS_PATH,
                BORDEAUX_PROMPTS_PATH, BORDEAUX_VOICES_PATH);
            return 0;
        }

#ifndef BUILD_DOCKER
        // start mqueue
        struct mq_attr attr{}, old_attr{};
        memset(&attr, 0, sizeof(attr));
        attr.mq_maxmsg = 10;
        attr.mq_msgsize = sizeof(ipc_event);
        auto mq = mq_open (BORDEAUX_IPCPATH, O_RDWR | O_CREAT, 0660, &attr);
        if(mq < 0)
            throw std::runtime_error(print_format("message queue failure {}", errno));
#endif

        // set miscellaneous configs
        if(is(opt_ipv6))
            ipv6 = true;

        if(is(opt_driver))
            driver_name = *opt_driver;

        if(::chdir(prefix_path))
            throw std::runtime_error(print_format("cannot access {}", prefix_path));
        // create default prompt workspaces, clean old session recordings....
        mkdir("prompts", 0770);
        mkdir("temp", 0700);
        server::dir temp("./temp");
        for(auto filename = "."s; !filename.empty(); filename = temp.get()) {
            if(filename[0] == '.')
                continue;

            filename = "temp/" + filename;
            if(filename.find('.') != std::string::npos)
                ::remove(filename.c_str());
        }

        if(getegid()) {
            umask(007); // FlawFinder: group write intended non root group
            auto pwd = getpwnam("nobody");
            if(pwd)
                server::nobody = pwd->pw_uid;
            endpwent();
        }
        else {
            umask(077); // FlawFinder: root write intended for root group
            nobody = getuid();
        }

        // initial logging
        logging = verbose.count();
#if !defined(BUILD_DOCKER) || defined(BUILD_SYSLOG)
        if(is(verbose))
            log_open(SERVER_ID, LOG_INFO);
        else
            log_open(SERVER_ID, LOG_NOTICE);
#endif

        // handlers for new and exit
        std::set_new_handler([]() {
            die(9, "error: out of memory\n");
        });

        // signal masks, we could add SIGUSR1..2 for special things
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGCHLD);
        sigaddset(&sigs, SIGTERM);
        sigaddset(&sigs, SIGUSR1);
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // bind driver
        if(!loader(BORDEAUX_DRIVERS_PATH, driver_name, "driver_"))
            die(-1, "error: driver {} failed: {}\n", driver_name, dlerror());

        // load plugins
#ifdef  BORDEAUX_VOICE_MODULES
        server::dir voices(BORDEAUX_VOICE_MODULES);
        for(auto filename = "."s; !filename.empty(); filename = voices.get()) {
        if(filename.find('.') == std::string::npos)
                loader(BORDEAUX_VOICE_MODULES, filename, "plugin_");
        }
#endif
        server::dir plugins(BORDEAUX_PLUGINS_PATH);
        for(auto filename = "."s; !filename.empty(); filename = plugins.get()) {
            if(filename[0] == '.')
                continue;

            if(begins_with(filename, "plugin_") && ends_with(filename, ".so")) {
                filename.erase(filename.size() - 3, 3);
                filename.erase(0, 7);
            }

            if(filename.find('.') != std::string::npos)
                loader(BORDEAUX_PLUGINS_PATH, filename, "plugin_");
        }

        // start running service
        auto config = reload();

        Driver::start_instance(config);
        Plugin::start(config);

        // create timers and other threads
        std::thread session_timers(&Session::start, 2);
        std::thread control_thread(&process_control, 1);
        std::this_thread::sleep_for(std::chrono::milliseconds(50));

#ifndef BUILD_DOCKER
        // system notifiers
        struct sigevent mq_event{};
        mq_event.sigev_notify = SIGEV_SIGNAL;
        mq_event.sigev_signo = SIGUSR1;
        mq_notify(mq, &mq_event);
        attr.mq_flags = O_NONBLOCK;
        mq_setattr(mq, &attr, &old_attr);
        ipc_event event{};
#endif
        auto pri = 0U;

        // handle signals while up
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                break;
            switch(sig) {
            case SIGCHLD:
                for(;;) {
                    auto status = 0;
                    auto pid = waitpid(-1, &status, WNOHANG);
                    if(pid == -1)
                        break;
                    proc::drop(pid, WEXITSTATUS(status));
                }
                break;
            case SIGINT:
                std::cout << '\n';
                stop();
                break;
            case SIGALRM:
                exit_code = 5;
                stop();
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                log_reset();
                log_info("reloading server");
                config = reload();
                Session::sync_store();
                Plugin::reload(config);
                Driver::reload_instance(config);
                break;
#ifndef BUILD_DOCKER
            case SIGUSR1:
                while(mq_receive(mq, reinterpret_cast<char *>(&event), sizeof(event), &pri) != -1)
                    process_events(event);
                mq_notify(mq, &mq_event);
                break;
#endif
            default:
                break;
            }
        }

        // normal teardown and exit
        session_timers.join();
        control_thread.join();
        proc::killall();    // just in case there are leftovers...
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            log_fail(-1, "{}", e.what());
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            log_fail(-1, "unknown failure");
        else
            die(-1, "error: unknown failure\n");
    }

    log_info("stopped server; reason={}", exit_code);
    return exit_code;
}

==> bordeaux_0.7.8.orig.tar.gz: bordeaux-0.7.8/server/main.cpp <==
// Copyright (C) 2020-2023 David Sugar, Tycho Softworks
// This code is licensed under MIT license

#include "main.hpp"
#include "utils.hpp"
#include "ipc.hpp"
#include "logger.hpp"
#include "network.hpp"
#include "drivers.hpp"
#include "sessions.hpp"

#include <args.hpp>
#include <keyfile.hpp>
#include <strings.hpp>
#include <filesystem.hpp>

#include <atomic>
#include <thread>
#include <chrono>
#include <fstream>
#include <iostream>
#include <csignal>
#include <climits>
#include <cstdio>
#include <new>

#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <pwd.h>
#include <mqueue.h>

#ifdef  SYSTEMD_SERVICE
#include <systemd/sd-daemon.h>
#else
inline auto sd_notify(int, const char *) {  // NOLINT
    return 0;
}
#endif

using namespace server;

std::atomic<bool> server::running = false;  // need not be atomic...but...
std::mutex server::loglock;
unsigned server::logging = 0;
bool server::ipv6 = false;

namespace {
auto driver_name = BORDEAUX_DEFAULT_DRIVER;
const char *config_file = SERVER_CONFIG_FILE;
const char *prefix_path = SERVER_PREFIX_PATH;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config(0,  "--config", "configuration file", "file");
args::option opt_driver('d',"--driver", "driver name to load", "path");
args::option helpflag('h',  "--help", "display this list");
#ifdef  DEBUG
args::option opt_module('m',"--module", "module path to load", "path");
#endif
args::option paths(0,       "--paths", "display paths used");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);

auto reload() {
    keyfile config({config_file, "custom.conf"});
    auto& hostname = config["server"]["hostname"];
    auto& username = config["server"]["username"];
    auto& media = config["server"]["media"];
    auto& language = config["server"]["language"];
    auto& location = config["server"]["location"];
    auto& voice = config["server"]["voice"];
    auto& ring_timer = config["timers"]["ring"];
    auto& answer_timer = config["timers"]["answer"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    if(hostname.empty()) {
        auto env = util::get_env("BORDEAUX_HOSTNAME", 255);
        if(!env || !*env)
            env = util::get_env("PUBLIC_HOSTNAME", 255);

        if(env && *env)
            hostname = env;
        else
            hostname = net::hostname();
    }

    if(language.empty()) {
        auto env = util::get_env("BORDEAUX_LANGUAGE", 16);
        if(!env || !*env)
            env = util::get_env("LANGUAGE", 16);

        if(!env || !*env)
            env = "en_US";

        language = env;
    }

    if(location.empty()) {
        auto env = util::get_env("BORDEAUX_LOCATION", 64);
        if(!env || !*env)
            env = util::get_env("LOCATION", 64);

        if(!env || !*env)
             env = "unknown";

        location = env;
    }

    if(voice.empty()) {
        auto env = util::get_env("BORDEAUX_VOICE", 64);
        if(!env || !*env)
            env = util::get_env("VOICE", 64);

        if(!env || !*env)
            env = "default";

        voice = env;
    }

    setenv("LANGUAGE", language.c_str(), 1);
    setenv("LOCATION", location.c_str(), 1);
    setenv("VOICE", voice.c_str(), 1);

    auto env = util::get_env("BORDEAUX_SESSIONS", 16);
    if(env && *env)
        config["server"]["limit"] = env;

    if(hostname.empty())
        hostname = net::hostname();

    if(ring_timer.empty())
        ring_timer = "4";

    if(answer_timer.empty())
        answer_timer = "30";

    if(!media.empty())
        media = net::uri_address(media);
    else
        media = net::uri_address(hostname);

    net::fqdn(hostname);

    if(username.empty()) {
        auto pwd = getpwuid(geteuid());
        if(pwd)
            username = pwd->pw_name;
        else
            username = "bordeaux";
        endpwent();
    }

    running = true;
    Plugin::config(config);
    return Driver::config_instance(config);
}

auto stop() {
    sd_notify(1, "STOPPING=1");
    running = false;
    std::this_thread::sleep_for(std::chrono::milliseconds(50));

    Driver::drop_instance();
    Plugin::stop();
    Session::stop();
    Driver::stop_instance();
    fsys::remove(SERVER_CONTROL_IPC);
}

void process_events(const ipc_event& event) {
    switch(event.type) {
    case IPC_STATUS:
        break;
    case IPC_LEVEL:
        logging = event.body.level;
        log_notice("logging level changed to {}", logging);
        break;
    case IPC_RELOAD:
        raise(SIGHUP);
        break;
    default:
        debug(4, "received unknown message {}", static_cast<int>(event.type));
    }
}

auto loader(const std::string& prefix, const std::string& name, const std::string& type) {
    auto module = type;
    module += name;
    module += ".so";
    auto dsopath = prefix + "/" + module;
    auto ptr = dlopen(dsopath.c_str(), RTLD_NOW | RTLD_GLOBAL);

#ifdef  DEBUG
    if(!ptr && name.find('/') != std::string::npos) {
        if(name[0] == '/')
            dsopath = name;
        else
            dsopath = ::getenv("PWD") + "/"s + name;    // FlawFinder: ignore
        ptr = dlopen(dsopath.c_str(), RTLD_NOW | RTLD_GLOBAL);
    }
#endif

    if(ptr)
        log_info("loading {}", dsopath);
    else
        log_error("{}: failed to load", name);
    return ptr != nullptr;
}
} // anon namespace

auto main(int argc, const char **argv) -> int { // NOLINT
#ifdef  DEBUG
    char pwd[512];
    getcwd(pwd, sizeof(pwd));
    setenv("PWD", pwd, 1);
    remove(SERVER_LOGGER_PATH "/bordeaux.log");
#endif

    auto exit_code = 0;
    log_reset();

    atexit([]{
        if(running) {
            stop();
            log_warn("exiting server");
        }
        mq_unlink(BORDEAUX_IPCPATH);
        shm_unlink(BORDEAUX_IPCPATH);
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                SERVER_ID " " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Bordeaux Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = *opt_prefix;

        if(is(version)) {
            print("Version: {}\n", PROJECT_VERSION);
            return 0;
        }

        if(is(paths)) {
            print("Config:  {}\nPrefix:  {}\nDrivers: {}\nPlugins: {}\nModules: {}\nScripts: {}\nPrompts: {}\nVoices:  {}\n",
                config_file, prefix_path,
                BORDEAUX_DRIVERS_PATH, BORDEAUX_PLUGINS_PATH,
                BORDEAUX_MODULES_PATH, BORDEAUX_SCRIPTS_PATH,
                BORDEAUX_PROMPTS_PATH, BORDEAUX_VOICES_PATH);
            return 0;
        }

        // start mqueue
        struct mq_attr attr{}, old_attr{};
        memset(&attr, 0, sizeof(attr));
        attr.mq_maxmsg = 10;
        attr.mq_msgsize = sizeof(ipc_event);
        auto mq = mq_open (BORDEAUX_IPCPATH, O_RDWR | O_CREAT, 0660, &attr);
        if(mq == MQD_INVALID)
            throw std::runtime_error(print_format("message queue failure {}", errno));

        // set miscellaneous configs
        if(is(opt_ipv6))
            ipv6 = true;

        if(is(opt_driver))
            driver_name = *opt_driver;

        umask(007);    // FlawFinder: group access even as root...

        // create default prompt workspaces, clean old session recordings....
        fsys::current_path(prefix_path);
        fsys::remove_all("temp");
        if(fsys::create_directory("prompts"))
            fsys::permissions("prompts", fsys::perms::owner_all | fsys::perms::group_read | fsys::perms::group_write | fsys::perms::group_exec);
        if(fsys::create_directory("temp"))
            fsys::permissions("temp", fsys::perms::owner_all);

        // initial logging
        logging = verbose.count();
        log_open(SERVER_ID, LOG_INFO);

        // handlers for new and exit
        std::set_new_handler([]() {
            die(9, "error: out of memory\n");
        });

        // signal masks, we could add SIGUSR1..2 for special things
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGCHLD);
        sigaddset(&sigs, SIGTERM);
        sigaddset(&sigs, SIGUSR1);
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // bind driver
        if(!loader(BORDEAUX_DRIVERS_PATH, driver_name, "driver_"))
            die(-1, "error: driver {} failed: {}\n", driver_name, dlerror());

        // load plugins
        scan_directory(BORDEAUX_PLUGINS_PATH, [](const auto& entry) {
            if (!entry.is_regular_file())
                return false;

            auto filename = entry.path().filename().string();
            if(begins_with(filename, "plugin_") && ends_with(filename, ".so")) {
                filename.erase(filename.size() - 3, 3);
                filename.erase(0, 7);
                loader(BORDEAUX_PLUGINS_PATH, filename, "plugin_");
                return true;
            }
            return false;
        });

#ifdef  DEBUG
        if(is(opt_module))
            loader("/noexist", *opt_module, "plugin_");
#endif

        // start running service
        auto config = reload();

        Driver::start_instance(config);
        Plugin::start(config);

        // create ipc info
        fsys::remove("ipc.json");
        std::ofstream ipc("ipc.json");
        ipc << R"({"ipc_path":")" << BORDEAUX_IPCPATH << R"(",)" << std::endl
            << R"( "event_size":)" << sizeof(ipc_event) << "," << std::endl
            << R"( "system_size":)" << sizeof(ipc_system) << "," << std::endl
            << R"( "session_size":)" << sizeof(ipc_session) << "," << std::endl
            << R"( "session_count":)" << Driver::instance()->limit() << std::endl
            << "}" << std::endl;
        ipc.close();

        // create timers and other threads
        std::thread session_timers(&Session::start, 2);
        std::this_thread::sleep_for(std::chrono::milliseconds(50));

        // system notifiers
        struct sigevent mq_event{};
        mq_event.sigev_notify = SIGEV_SIGNAL;
        mq_event.sigev_signo = SIGUSR1;
        mq_notify(mq, &mq_event);
        attr.mq_flags = O_NONBLOCK;
        mq_setattr(mq, &attr, &old_attr);
        ipc_event event{};
        auto pri = 0U;

        // handle signals while up
        auto timer = sync_clock() + std::chrono::seconds(60);
        if(!eq(driver_name, "test"))
            alarm(60);
        log_info("started server");
        sd_notify(0, "READY=1");
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                break;
            switch(sig) {
            case SIGCHLD:
                for(;;) {
                    auto status = 0;
                    auto pid = waitpid(-1, &status, WNOHANG);
                    if(pid == -1)
                        break;
                }
                break;
            case SIGINT:
                std::cout << '\n';
                stop();
                break;
            case SIGALRM:
                // abort if testing..
                if(eq(driver_name, "test")) {
                    exit_code = 5;
                    stop();
                }
                // sync alarm every 60 seconds, used for stats...
                sd_notify(0, "WATCHDOG=1");
                debug(3, "minute timer wakeup");
                timer += std::chrono::seconds(60);
                alarm(std::chrono::duration_cast<std::chrono::seconds>(timer - sync_clock()).count());
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                sd_notify(0, "RELOADING=1");
                log_reset();
                log_info("reloading server");
                config = reload();
                Session::sync_store();
                Plugin::reload(config);
                Driver::reload_instance(config);
                sd_notify(0, "READY=1");
                break;
            case SIGUSR1:
                while(mq_receive(mq, reinterpret_cast<char *>(&event), sizeof(event), &pri) != -1)
                    process_events(event);
                mq_notify(mq, &mq_event);
                break;
            default:
                break;
            }
        }

        // normal teardown and exit
        session_timers.join();
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            log_fail(-1, "{}", e.what());
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            log_fail(-1, "unknown failure");
        else
            die(-1, "error: unknown failure\n");
    }

    log_info("stopped server; reason={}", exit_code);
    return exit_code;
}
